from . import subject

class Store():
    """
    Stores all the subjects as a dictionary, with their keys as the index
    """
    def __init__(self):
        self.subjects = {}

    # getters
    def get_subjects(self):
        """ Returns a tuple with the indices of all subjects"""
        return tuple(self.subjects.keys()) 
    
    def get_subject_by_index(self, index):
        """ Takes subject-index as the input and returns the corresponding subject """
        try:
            return self.subjects[index]
        except:
            return None
    
    def get_plotables(self, index):
        """ Gets all the time-series that can be plotted, for a particular subject"""
        return self.subjects[index].get_plotables()
    
    def get_data_labels(self, index):
        """ Gets the labels for all the time-series data for that subject"""
        return self.subjects[index].get_data_labels()
    
    # misc.
    def get_time_series(self, s):
        """ Gets the time-series object corresponding to s"""
        s       = s.split("_")
        index   = s[0]
        ts_type = s[1]
        ts_num  = int(s[2])
        return self.subjects[index].get_time_series_by_index(ts_type, ts_num)
    
    def get_info(self, s):
        """ Gets all the information about a time-series to be displayed to the logging-window """
        ts            = self.get_time_series(s)
        subject_index = ts.get_subject_index()
        subject       = self.subjects[subject_index]
        dic           = ts.get_info()
        if ts.get_label()   == "BOLD":
            dic["Input File"]          = ts.get_inputfile()
        elif ts.get_label() == "Preprocessed-BOLD":
            dic["Mask File"]           = ts.get_maskfile()
            dic["Associated Raw BOLD"] = subject_index + "_BOLD_"               + subject.get_time_series_pos(ts.get_BOLD_Raw())
        elif ts.get_label() == "HRF":
            dic["Associated BOLD"]     = subject_index + "_Preprocessed-BOLD_"  + subject.get_time_series_pos(ts.get_associated_BOLD())
        elif ts.get_label() == "Deconvolved-BOLD":            
            dic["Associated HRF"]      = subject_index + "_HRF_"                + subject.get_time_series_pos(ts.get_associated_HRF())
        return dic

    def is_present(self, subject_index):
        """ Checks whether a subject-index is present is already present"""
        if subject_index in self.subjects.keys():
            return True
        return False
    
    def add_subject(self, sub):
        """ Adds a new subject """
        self.subjects[sub.get_subject_index()] = sub
    
    def remove_subject(self, sub):
        """ Removing a subject """
        try:
            del self.subjects[sub.get_subject_index()]
        except:
            return 
    
    def number_of_subjects(self):
        """ Gets the number of subjects """
        return len(self.subjects.keys())
    
    def save_info(self, s, out):
        """ Saves all the time-series objects for a particular subject in the form of .mat files """
        ts = self.get_time_series(s)
        return ts.save_info(out+"/sub-"+s+".mat")


from copy import deepcopy
from tkinter import Toplevel, Button, Entry, Label, DISABLED

class ParameterWindow():
    def __init__(self):
        self.window = Toplevel()
        self.window.title("Parameters")
        self.parameters = {}
        self.labels     = []
        self.entries    = []
        # get screen width and height
        screen_width  = self.window.winfo_screenwidth()
        screen_height = self.window.winfo_screenheight()
        self.window.geometry("350x420+%d+%d" % (screen_width*(280.0/1900), (((1040.0-220)/1000)*screen_height)-390))
        
    def updateParameters(self):
        for i in range(len(self.labels)):
            self.parameters[self.labels[i].cget('text')] = self.entries[i].get()
        parameters = dict(self.parameters)
        return deepcopy(parameters)

    def getParameters(self):
        return deepcopy(self.parameters)
    
    def setParameters(self, dic):
        self.parameters = deepcopy(dic) 
    
    def display(self): 
        # removing existing widgets
        for each in self.labels:
            each.destroy()
        for each in self.entries:
            each.destroy()
        # making new widgets
        self.labels  = []
        self.entries = []
        for each in self.parameters.keys():
            self.labels.append(Label(self.window, text=each))
            self.entries.append(Entry(self.window, width=15))
            if each == "passband" or each == "passband_deconvolve":
                self.entries[-1].insert(0, str(self.parameters[each][0]) + "," + str(self.parameters[each][1]))
            else:
                self.entries[-1].insert(0, str(self.parameters[each]))
            if each == "lag" or each == "dt":
                self.entries[-1].configure(state=DISABLED)
        row_i = 0
        while row_i < len(self.labels):
            key = self.labels[row_i].cget("text")
            if key == "estimation" or key == "temporal_mask":
                row_i += 1
                continue
            self.labels[row_i].grid(row=row_i,column=0,padx=(5,5),pady=(5,5))
            self.entries[row_i].grid(row=row_i,column=1,padx=(5,5),pady=(5,5))
            row_i += 1



import math
import numpy as np
from scipy.stats import gamma
from rsHRF import canon, spm_dep, sFIR

import warnings
warnings.filterwarnings("ignore")

"""
BASIS FUNCTION COMPUTATION
"""

def get_basis_function(bold_sig_shape, para):
    N, nvar = bold_sig_shape
    dt = para['dt'] # 'time bin for basis functions {secs}';
    l  = para['len']
    if "gamma" in para['estimation']:
        pst = np.arange(0,l+0.01,dt) #the 0.01 is because the rounding is different between python and matlab
        bf  = gamma_bf(pst,para['order'])
    elif 'fourier' in para['estimation'] or 'hanning' in para['estimation']:
        pst = np.arange(0,l+0.01,dt) #the 0.01 is because the rounding is different between python and matlab
        pst = pst/max(pst)
        bf  = fourier_bf(pst,para)
    elif 'canon' in para['estimation']:
        bf = canon2dd_bf(bold_sig_shape, para)
    bf = spm_dep.spm.spm_orth(np.asarray(bf))
    return bf

def canon2dd_bf(data_shape, xBF):
    """
    Returns canon basis functions
    """
    N, nvar = data_shape
    bf = canon.canon_hrf2dd.wgr_spm_get_canonhrf(xBF)
    if 'Volterra' in xBF:
        if xBF['Volterra'] == 2:
            bf2 = np.einsum('i...,j...->ij...',bf.T,bf.T).reshape(-1, bf.shape[0]).T
            bf  = np.column_stack((bf, bf2))
    return bf

def fourier_bf(pst,para):
    """
    Returns Fourier (Hanning) basis functions
    """
    if "hanning" in para['estimation']:
        g = (1 - np.cos(2*math.pi*pst)) / 2
    else:
        g = np.ones(len(pst))
    sin_ = lambda x : np.sin(x*2*math.pi)
    cos_ = lambda x : np.cos(x*2*math.pi) 
    sin_ = np.vectorize(sin_)
    cos_ = np.vectorize(cos_)
    arr = np.arange(1, para['order']+1)
    s = sin_(np.einsum('i,j->ij',arr,pst))
    c = cos_(np.einsum('i,j->ij',arr,pst))
    s = np.multiply(g,s)
    c = np.multiply(g,c)
    g = np.expand_dims(g, axis=1).T
    return np.concatenate((g, s, c), axis=0).T

def gamma_bf(u,h):
    """
    Returns Gamma basis functions
    """
    arr = np.arange(2, h+2)
    f = np.vectorize(gamma.pdf, signature='(n),()->(n)')
    m = np.power(2, arr)
    return f(u,m).T




class Status():
    def __init__(self, state, error="", info="", time="", dic={}):
        self.state = state
        self.info  = info
        self.error = error
        self.time  = time
        self.dic   = dic

    # setters
    def set_state(self, b):
        self.state = b

    def set_info(self, s):
        self.info  = s
    
    def set_error(self, e):
        self.error = e
    
    def set_time(self, t):
        self.time  = t
    
    def set_dic(self, d):
        self.dic   = d
    
    # getters
    def get_state(self):
        return self.state 
    
    def get_info(self):
        return self.info 
    
    def get_error(self):
        return self.error
    
    def get_time(self):
        return self.time
    
    def get_dic(self):
        return self.dic


import os
from tkinter       import Toplevel, Checkbutton, IntVar, Button, filedialog, NORMAL, DISABLED, OptionMenu, StringVar, Label


class InputWindow():
    def __init__(self):
        # input window
        window = Toplevel()
        window.title("Input Window")
        # get screen width and height
        screen_width  = window.winfo_screenwidth()
        screen_height = window.winfo_screenheight()
        # placing the toplevel
        window.geometry("350x220+%d+%d" % ((280/1900)*screen_width, ((((1040.0-220)/1000)*screen_height)-390)-280))
        # variables which shall get sent to the front end
        self.input_file = ()
        self.mask_file  = ()
        self.file_type  = ()
        self.output_dir = ()
        # other class vairables
        # 1 corresponds to BIDS input
        self.inputFormatVar = IntVar()
        self.inputFormatVar.set(0) 
        # 1 corresponds to mask file being present in the BIDS directory
        self.maskFormatVar = IntVar()
        self.maskFormatVar.set(0) 
        # selecting the estimation rule
        self.estimationOption = StringVar()
        self.estimationOption.set('canon2dd')

        def getInputFile():
            if self.inputFormatVar.get(): # input takes a directory
                self.input_file = filedialog.askdirectory(initialdir=os.getcwd())
                maskFormat.configure(state=NORMAL)
            else:
                self.input_file = filedialog.askopenfilename(initialdir=os.getcwd(), title="Input File Path", filetypes=(("nifti files", "*.nii"), ("nifti files", "*.nii.gz"), ("gifti files", "*.gii"), ("gifti files", "*.gii.gz")))
                maskFormat.configure(state=DISABLED)
                try:
                    self.file_type = os.path.splitext(self.input_file)[1]
                except: 
                    self.file_type = ()             
            try: 
                inputFileLabel.configure(text=self.input_file.split('/')[-1])
            except:
                inputFileLabel.configure(text="")

        def maskFormatButtonState():
            if self.maskFormatVar.get():
                maskFormatButton.configure(state=DISABLED)
            else:
                maskFormatButton.configure(state=NORMAL)

        def inputFormatButtonState():
            if self.inputFormatVar.get():
                maskFormat.configure(state=NORMAL)
            else:
                maskFormat.configure(state=DISABLED)
            maskFormatButtonState()

        def getMaskFile():
            self.mask_file = filedialog.askopenfilename(initialdir=os.getcwd(), title="Input File Path", filetypes=(("nifti files", "*.nii"), ("nifti files", "*.nii.gz"), ("gifti files", "*.gii"), ("gifti files", "*.gii.gz")))
            try:
                maskFileLabel.configure(text=self.mask_file.split("/")[-1])
            except:
                maskFileLabel.configure(text="")

        def getOutputDir():
            self.output_dir = filedialog.askdirectory(initialdir=os.getcwd())
            try:
                outputPathLabel.configure(text="Output path: " + self.output_dir.split("/")[-1])
            except:
                outputPathLabel.configure(text="")

        # defining widgets
        inputFormat        = Checkbutton(window, text="BIDS Format",       variable=self.inputFormatVar,                command=inputFormatButtonState)
        maskFormat         = Checkbutton(window, text="Mask File in BIDS", variable=self.maskFormatVar, state=DISABLED, command=maskFormatButtonState)
        inputFormatButton  = Button     (window, text="Select Input",                                                   command=getInputFile, height=1, width=20)
        maskFormatButton   = Button     (window, text="Select Mask File",                               state=NORMAL,   command=getMaskFile,  height=1, width=20)
        outputPathButton   = Button     (window, text="Select Output Directory",                                        command=getOutputDir, height=1, width=20)
        estimationLabel    = Label      (window, text="Estimation Rule: ")
        inputFileLabel     = Label      (window, text="")
        maskFileLabel      = Label      (window, text="")
        outputPathLabel    = Label      (window, text="")
        estimationDropDown = OptionMenu (window, self.estimationOption, "canon2dd", "sFIR", "FIR", "gamma", "fourier", "fourier w/ hanning")
        # placing widgets
        inputFormat.grid       (row=0,column=0,padx=(5,5),pady=(5,5))
        inputFormatButton.grid (row=0,column=1,padx=(5,5),pady=(5,5))
        inputFileLabel.grid    (row=1,column=0,padx=(5,5),pady=(5,5),columnspan=2)
        maskFormat.grid        (row=2,column=0,padx=(5,5),pady=(5,5))
        maskFormatButton.grid  (row=2,column=1,padx=(5,5),pady=(5,5))
        maskFileLabel.grid     (row=3,column=0,padx=(5,5),pady=(5,5),columnspan=2)
        outputPathButton.grid  (row=4,column=1,padx=(5,5),pady=(5,5))
        outputPathLabel.grid   (row=4,column=0,padx=(5,5),pady=(5,5))
        estimationLabel.grid   (row=5,column=0,padx=(5,5),pady=(5,5))
        estimationDropDown.grid(row=5,column=1,padx=(5,5),pady=(5,5))

    def getInput(self):
        if self.inputFormatVar.get() * self.maskFormatVar.get():
            mode = "bids"
        elif self.inputFormatVar.get():
            mode = "bids w/ atlas"
        else:
            mode = "file"
        return (self.input_file, self.mask_file, self.file_type, mode, self.estimationOption.get(), self.output_dir)


import sys
import numpy as np
from copy           import deepcopy
from ...misc.status import Status

class Parameters():
    """ 
    rsHRF-Toolbox Parameters
    For more information, visit https://github.com/BIDS-Apps/rsHRF/tree/master/rsHRF
    """
    def __init__(self):
        # initialize default parameters
        self.estimation          = 'canon2dd'
        self.passband            = [0.01, 0.08]
        self.passband_deconvolve = [0.0, sys.float_info.max]
        self.TR                  = 2.0 
        self.localK              = 1
        self.T                   = 3
        self.T0                  = 1
        self.TD_DD               = 2
        self.AR_lag              = 1
        self.thr                 = 1
        self.order               = 3
        self.volterra            = 0
        self.len                 = 24      
        self.temporal_mask       = []
        self.min_onset_search    = 4
        self.max_onset_search    = 8
        self.dt                  = self.TR/self.T
        self.lag                 = np.arange(np.fix(self.min_onset_search / self.dt),
                                   np.fix(self.max_onset_search / self.dt) + 1,
                                   dtype='int')

    # getters
    def get_estimation(self):
        return self.estimation 
    
    def get_passband(self):
        return deepcopy(self.passband)

    def get_passband_deconvolve(self):
        return deepcopy(self.passband_deconvolve)
    
    def get_TR(self):
        return self.TR
    
    def get_localK(self):
        return self.localK
    
    def get_T(self):
        return self.T 
    
    def get_T0(self):
        return self.get_T0
    
    def get_TD_DD(self):
        # TD_DD is only relevant for canon2dd estimation
        if self.estimation == 'canon2dd':
            return self.TD_DD 
        else:
            return None 
    
    def get_AR_lag(self):
        return self.get_AR_lag 
    
    def get_thr(self):
        return self.thr 
    
    def get_order(self):
        # order is only relevant for fourier and gamma estimation
        if 'gamma' in self.get_estimation or 'fourier' in self.get_estimation:
            return self.order 
        else:
            return None 
    
    def get_Volterra(self):
        # volterra is only relevant for canon2dd estimation
        if self.get_estimation == 'canon2dd':
            return self.volterra
        else:
            return None 
    
    def get_len(self):
        return self.len 
    
    def get_temporal_mask(self):
        return tuple(self.temporal_mask)
    
    def get_min_onset_search(self):
        return self.min_onset_search
    
    def get_max_onset_search(self):
        return self.max_onset_search
    
    # setters
    def set_estimation(self,s):
        self.estimation = s
        return Status(True)
    
    def set_passband(self, l):
        """ 
        Takes a list (with two entries) as input and sets the 
            passband-range
        Checks whether both the values are non-negative 
        """
        try:
            l = [float(i) for i in l.split(",")]
            if l[0] < 0 or l[1] < 0:
                return Status(False, error="Passband frequency values cannot be negative")
        except:
            return Status(False, error="Bad datatype for passband range")
        else:
                self.passband = deepcopy(l)
                return Status(True)
    
    def set_passband_deconvolve(self, l):
        """ 
        Takes a list (with two entries) as input and sets the 
            passband-range
        Checks whether both the values are non-negative 
        """
        try:
            l = [float(i) for i in l.split(",")]
            if l[0] < 0 or l[1] < 0:
                return Status(False, error="Passband frequency values cannot be negative")
        except:
            return Status(False, error="Bad datatype for passband range")
        else:
                self.passband_deconvolve = deepcopy(l)
                return Status(True)
    
    def set_TR(self, TR):
        """ 
        Checks whether the value is postiive
        and updates 'dt' as it is dependent on TR
        """
        try:
            TR = float(TR)
        except:
            return Status(False,error="Bad Datatype For TR")
        else:
            if TR <= 0:
                return Status(False, error="BOLD Repetition Time must be greater than 0")
            else:
                self.TR = TR
                self.update_dt()
                return Status(True)
    
    def set_localK(self, localK):
        """
        Checks whether the value of localK is positive
        """
        try:
            localK = int(localK)
        except:
            return Status(False,error="Bad Datatype For localK")
        else:
            if localK <= 0:
                return Status(False, error="localK must be greater than 0")
            else:
                self.localK = localK
                return Status(True)
    
    def set_T(self, T):
        """
        Checks whether the value of T >= 1
        and updates 'dt' as it is dependent on T
        """
        try:
            T = int(T)
        except:
            return Status(False, error="Bad Datatype For T")
        else:
            if T < 1:
                return Status(False, error="Magnification factor must not be less than 1") 
            else:
                self.T = T 
                self.update_dt()
                return Status(True)
    
    def set_T0(self, T0):
        try:
            T0 = int(T0)
        except:
            return Status(False, error="Bad Datatype For T0")
        else:
            self.T0 = T0 
            return Status(True)
    
    def set_TD_DD(self, TD_DD):
        """
        Sets the value of TD_DD if the estimation is canon2dd
        """
        if self.estimation == 'canon':
            try:
                TD_DD = int(TD_DD)
            except:
                return Status(False, error="Magnification factor must not be less than 1") 
            if TD_DD not in [0,1,2]:
                return Status(False,error="TD_DD can only take one of these values: 0, 1, 2")
            else:
                self.TD_DD = TD_DD 
                return Status(True)
        else:
            return Status(True)
    
    def set_AR_lag(self, AR_lag):
        """
        Checks whether the AR_lag is non-negative
        """
        try:
            AR_lag = int(AR_lag)
        except:
            return Status(False, error="Bad datatype for AR_lag")
        else:
            if AR_lag < 0:
                return Status(False, error="AR_lag must not be negative")
            else:
                self.AR_lag = AR_lag 
                return Status(True)
    
    def set_thr(self, thr):
        """
        If estimation is FIR or sFIR, thr is a list
        Otherwise, it is an int
        """
        if 'FIR' not in self.estimation :
            try:
                thr = int(thr)
            except:
                return Status(False, error="Bad datatype for thr")
            else:
                self.thr = thr 
                return Status(True)
        else:
            try:
                thr = [int(i) for i in thr.split(",")]
            except:
                return Status(False, error="Bad datatype for thr")
            else:
                self.thr = thr 
                return Status(True)
    
    def set_order(self, order):
        """
        Checks whether the order lies between 1 to 60
        """
        try:
            order = int(order)
        except:
            return Status(False, error="Bad datatype for order")
        else:
            if order < 0 or order > 60:
                status = Status(False)
                if order < 0:
                    status.set_error("Order must not be negative")
                elif order > 60:
                    status.set_error("Order is too high")
                return status
            else:
                self.order = order 
                return Status(True)
    
    def set_Volterra(self, volterra):
        """
        Sets Volterra if the estimation rule is canon2dd
        """
        if self.estimation == 'canon':
            try:
                volterra = int(volterra)
            except:
                return Status(False, error="Bad datatype for Volterra")
            else:
                self.volterra = volterra
                return Status(True)
        else:
            return Status(True)
    
    def set_len(self, length):
        """
        Checks whether the length of the hemodynamic response function is non-negative
        """
        try:
            length = int(length)
        except:
            return Status(False, error="Bad datatype for len")
        else:
            if length <= 0:
                status = Status(False)
                status.set_error("HRF length must not be postitive")
                return status
            else:
                self.len = length 
                return Status(True)
    
    def set_temporal_mask(self, temporal_mask):
        self.temporal_mask = deepcopy(temporal_mask)
        return Status(True)
    
    def set_min_onset_search(self, mos):
        """
        Checks whether the min-onset-search is smaller than max-onset-search
        and non-negative
        """
        try:
            mos = int(mos)
        except:
            return Status(False, error="Bad datatype for min_onset_search")
        else:
            if mos > self.max_onset_search or mos < 0:
                status = Status(False)
                if mos > self.max_onset_search:
                    status.set_error("Min Onset Search must not be greater than Max Onset Search")
                elif mos < 0:
                    status.set_error("Onset Search values must not be negative")
                return status 
            else:
                self.min_onset_search = mos
                self.update_lag()
                return Status(True) 
    
    def set_max_onset_search(self, mos):
        """
        Checks whether max-onset-search is greater than min-onset-search
        and non-negative
        """
        try:
            mos = int(mos)
        except:
            return Status(False, error="Bad datatype for max_onset_search")
        else:
            if mos < self.min_onset_search or mos < 0:
                status = Status(False)
                if mos < self.min_onset_search:
                    status.set_error("Max Onset Search must not be lesser than Min Onset Search")
                elif mos < 0:
                    status.set_error("Onset Search values must not be negative")
                return status 
            else:
                self.max_onset_search = mos
                self.update_lag()
                return Status(True)

    def update_dt(self):
        """
        Re-calculating dt
        """
        self.dt  = self.TR / self.T 
        self.update_lag()
    
    def update_lag(self):
        """
        Re-calculating lag
        """
        self.lag = np.arange(np.fix(self.min_onset_search / self.dt),
                    np.fix(self.max_onset_search / self.dt) + 1,
                    dtype='int')
    
    def get_parameters(self):
        """
        Gets all the parameters in the form of a dictionary for rsHRF computation
        """
        para                            = {}
        para['estimation']              = self.estimation
        para['passband']                = deepcopy(self.passband)
        para['passband_deconvolve']     = deepcopy(self.passband_deconvolve)
        para['TR']                      = self.TR
        para['T']                       = self.T 
        para['localK']                  = self.localK
        para['T0']                      = self.T0
        para['AR_lag']                  = self.AR_lag
        para['thr']                     = self.thr
        para['len']                     = self.len
        para['temporal_mask']           = deepcopy(self.temporal_mask)
        para['min_onset_search']        = self.min_onset_search
        para['max_onset_search']        = self.max_onset_search
        if self.estimation == 'canon2dd':
            para['TD_DD']               = self.TD_DD
            para['Volterra']            = self.volterra
        elif 'gamma' in self.estimation or 'fourier' in self.estimation:
            para['order']               = self.order 
        para['dt']                      = self.dt 
        para['lag']                     = self.lag 
        return para
        
    def set_parameters(self, dic):
        """
        Takes a dictionary as input and sets all the rsHRF parameters accordingly
        """
        for key in dic.keys():
            if key   == "estimation":
                out  = self.set_estimation(dic[key])
            elif key == "passband":
                out  = self.set_passband(dic[key])
            elif key == "passband_deconvolve":
                out  = self.set_passband_deconvolve(dic[key])
            elif key == "T":
                out  = self.set_T(dic[key])
            elif key == "TR":
                out  = self.set_TR(dic[key])
            elif key == "localK":
                out  = self.set_localK(dic[key])
            elif key == "T0":
                out  = self.set_T0(dic[key])
            elif key == "TD_DD":
                out  = self.set_TD_DD(dic[key])
            elif key == "AR_lag":
                out  = self.set_AR_lag(dic[key])
            elif key == "thr":
                out  = self.set_thr(dic[key])
            elif key == "order":
                out  = self.set_order(dic[key])
            elif key == "Volterra":
                out  = self.set_thr(dic[key])
            elif key == "len":
                out  = self.set_len(dic[key])
            elif key == "temporal_mask":
                out  = self.set_temporal_mask(dic[key])
            elif key == "min_onset_search":
                out  = self.set_min_onset_search(dic[key])
            elif key == "max_onset_search":
                out  = self.set_max_onset_search(dic[key])
            if not out.get_state():
                return out 
        return Status(True, info="Parameters Updated Succefully")
            
    def compareParameters(self, p):
        """
        Takes another parameter object and determines if it is equal 
        to the this object
        """
        x = self.get_parameters()
        y = p.get_parameters()
        for key in x.keys():
            if key not in y.keys():
                return False 
            elif key == "dt" or key == "lag":
                continue 
            else:
                if x[key] != y[key]:
                    return False 
        return True



import numpy as np
from scipy.io   import savemat
from copy       import deepcopy

from .hrf                         import HRF 
from .timeseries                  import TimeSeries
from ...datatypes.misc.parameters import Parameters

class Bold_Deconv(TimeSeries):
    """
    This stores the Deconvolved BOLD Time-series
    
    Attributes:
        1. HRF          : The HRF used to obtain the Deconvolved BOLD
        2. event_num    : The event-numbers
    """
    def __init__(self, label="",ts=np.array([]),subject_index="", para=Parameters()):
        TimeSeries.__init__(self, label="",ts=np.array([]),subject_index="", para=Parameters())
        self.label         = label
        self.subject_index = subject_index
        self.timeseries    = ts
        self.shape         = ts.shape
        self.parameters    = deepcopy(para) 
        self.HRF           = HRF()
        self.event_num     = np.array([])
    
    #setters
    def set_HRF(self, HRF):
        self.HRF = HRF 
    def set_event_num(self, ev):
        self.event_num = ev 
    
    #getters
    def get_event_num(self):
        return self.event_num
    def get_associated_HRF(self):
        return self.HRF 
    
    #misc
    def compareTimeSeries(self, ts):
        """ Compares another time-series with itself to determine if both are identical 
            Two checks are performed:
                1. Label
                2. HRF
            If all the three comparisions return true, then both the HRF
                time-series objects are identical
        """
        if self.label == ts.get_label() \
            and self.HRF.compareTimeSeries(ts.get_associated_HRF()):
            return True 
        else:
            return False
    def save_info(self, name):
        """ Saves the information about the time-series in a .mat file """
        dic = {}
        dic["timeseries"] = self.timeseries
        dic["eventNum"]   = self.event_num
        para              = self.parameters.get_parameters()
        for each in para.keys():
            dic[each] = para[each] 
        savemat(name, dic)
        return True


import numpy as np
from scipy        import linalg
from ..processing import knee

import warnings
warnings.filterwarnings("ignore")

def wgr_regress(y, X):
    n, ncolX = X.shape
    Q,R,perm = linalg.qr(X, mode='economic', pivoting=True)
    if R.ndim == 0:
        p = 0
    elif R.ndim == 1:
        p = int(abs(R[0]) > 0)
    else:
        if np.amin(R.shape) == 1:
            p = int(abs(R[0]) > 0)
        else:
            p = np.sum(np.abs(np.diagonal(R)) > abs(max(n, ncolX)*np.spacing(R[0][0])))
    if p < ncolX:
        R = R[0:p,0:p]
        Q = Q[:,0:p]
        perm = perm[0:p]
    b = np.zeros((ncolX))
    if(R.shape[0] == R.shape[1]):
        try:
            b[perm] = linalg.solve(R,np.matmul(Q.T,y))
        except:
            b[perm] = linalg.lstsq(R,np.matmul(Q.T,y))
    else:
        b[perm] = linalg.lstsq(R,np.matmul(Q.T,y))
    return b

def wgr_glsco(X, Y, sMRI = [], AR_lag=0, max_iter=20):
    """
    Linear regression when disturbance terms follow AR(p)
    -----------------------------------
    Model:
    Yt = Xt * Beta + ut ,
    ut = Phi1 * u(t-1) + ... + Phip * u(t-p) + et
    where et ~ N(0,s^2)
    -----------------------------------
    Algorithm:
    Cochrane-Orcutt iterated regression (Feasible generalized least squares)
    -----------------------------------
    Usage:
    Y = dependent variable (n * 1 vector)
    X = regressors (n * k matrix)
    AR_lag = number of lags in AR process
    -----------------------------------
    Returns:
    Beta = estimator corresponding to the k regressors
    """
    nobs, nvar = X.shape
    if sMRI == []:
        Beta = wgr_regress(Y,X)
    else:
        sMRI = np.array(sMRI)
        try:
            Beta = linalg.solve(np.matmul(X.T,X)+sMRI,np.matmul(X.T,Y))
        except:
            Beta = linalg.lstsq(np.matmul(X.T,X)+sMRI,np.matmul(X.T,Y))
    resid = Y - np.matmul(X,Beta)
    if AR_lag == 0:
        res_sum = np.cov(resid)
        return res_sum, Beta
    max_tol = min(1e-6, max(np.absolute(Beta)) / 1000)
    for r in range(max_iter):
        Beta_temp = Beta
        X_AR = np.zeros((nobs - (2 * AR_lag), AR_lag))
        for m in range(AR_lag):
            X_AR[:, m] = resid[AR_lag - m - 1:nobs - AR_lag - m - 1]
        Y_AR = resid[AR_lag:nobs - AR_lag]
        AR_para = wgr_regress(Y_AR, X_AR)
        X_main = X[AR_lag:nobs, :]
        Y_main = Y[AR_lag:nobs]
        for m in range(AR_lag):
            X_main = \
                X_main - (AR_para[m] * (X[AR_lag - m - 1:nobs - m - 1, :]))
            Y_main = Y_main - (AR_para[m] * (Y[AR_lag - m - 1:nobs - m - 1]))
        if sMRI == []:
            Beta = wgr_regress(Y_main, X_main)
        else:
            try:
                Beta = linalg.solve(np.matmul(X_main.T,X_main)+sMRI,np.matmul(X_main.T,Y_main))
            except:
                Beta = linalg.lstsq(np.matmul(X_main.T,X_main)+sMRI,np.matmul(X_main.T,Y_main))
        resid = Y[AR_lag:nobs] - X[AR_lag:nobs, :].dot(Beta)
        if(max(np.absolute(Beta - Beta_temp)) < max_tol):
            break
    res_sum = np.cov(resid)
    return res_sum, Beta
    
def Fit_sFIR2(output, length, TR, input, T, flag_sfir, AR_lag):
    NN = int(np.floor(length/TR))
    _input = np.expand_dims(input[0], axis=0)
    X = linalg.toeplitz(input, np.concatenate((_input, np.zeros((1, NN-1))), axis = 1))
    X = np.concatenate((X, np.ones((input.shape))), axis = 1)
    if flag_sfir:
        fwhm = 7 #fwhm=7 seconds smoothing - ref. Goutte
        nh = NN-1
        dt = TR
        _ = np.expand_dims(np.arange(1, nh+1).T, axis=1)
        C = np.matmul(_,np.ones((1, nh)))
        h = np.sqrt(1./(fwhm/dt))
        v = 0.1
        R = v * np.exp(-h/2 * (C-C.T)**2)
        RI = linalg.inv(R)
        MRI = np.zeros((nh + 1, nh + 1))
        MRI[0:nh,0:nh] = RI;
        sigma = 1
        sMRI0 = sigma**2*MRI
        sMRI = np.zeros((NN+1, NN+1))
        sMRI[0:NN,0:NN] = sMRI0; 
        if AR_lag == 0:
            try:
                hrf = linalg.solve((np.matmul(X.T,X)+sMRI),np.matmul(X.T,output))
            except:
                hrf = linalg.lstsq((np.matmul(X.T,X)+sMRI),np.matmul(X.T,output))
            resid = output - np.matmul(X, hrf)
            res_sum = np.cov(resid)
        else:
            res_sum, hrf = wgr_glsco(X,output,AR_lag=AR_lag,sMRI=sMRI);
    else:
        if AR_lag == 0:
            hrf = linalg.lstsq(X,output)
            hrf = hrf[0]
            resid = output - np.matmul(X, hrf)
            res_sum = np.cov(resid)
        else:
            res_sum, hrf = wgr_glsco(X,output,AR_lag=AR_lag)
    return hrf, res_sum

def wgr_FIR_estimation_HRF(u, dat, para, N):
    if para['estimation'] == 'sFIR':
        firmode = 1
    else:
        firmode = 0
    lag = para['lag']
    nlag = np.amax(lag.shape)
    len_bin = int(np.floor(para['len'] / para['TR']))
    hrf = np.zeros((len_bin+1, nlag))
    Cov_E = np.zeros((1, nlag))
    kk = 0
    for i_lag in range(1, nlag + 1):
        RR = u - i_lag
        RR = RR[RR >= 0]
        if RR.size != 0:
            design = np.zeros((N, 1))
            design[RR] = 1
            hrf_kk, e3 = Fit_sFIR2(dat, para['len'], para['TR'], design, len_bin, firmode, para['AR_lag'])
            hrf[:, kk] = np.ravel(hrf_kk)
            Cov_E[:, kk] = (np.ravel(e3))
        else:
            Cov_E[:, kk] = np.inf
        kk += 1
    placeholder, ind = knee.knee_pt(np.ravel(Cov_E))
    if ind == np.amax(Cov_E.shape) - 1:
        ind = ind - 1
    rsH = hrf[:,ind+1]
    return rsH, u


import rsHRF.spm_dep
import rsHRF.processing
import rsHRF.canon
import rsHRF.sFIR
import rsHRF.parameters
import rsHRF.fourD_rsHRF
import rsHRF.CLI
__all__ = ["spm_dep", "processing", "canon", "utils",
           "sFIR", "parameters", "basis_functions", "fourD_rsHRF.py", "CLI.py"]



import numpy as np
from scipy.io   import savemat
from copy       import deepcopy

from ...datatypes.misc.parameters import Parameters
from .timeseries                  import TimeSeries

class HRF(TimeSeries):
    """
    This stores the Hemodynamic Response Function Time-series
    
    Attributes:
        1. BOLD       = stores the associated Preprocessed-BOLD time-series object through which it was retrieved
        2. PARA       = stores the HRF parameters (Full-width at half-max, Time-to-peak and height)
        3. event_bold = stores the bold-events
    """
    def __init__(self, label="",ts=np.array([]),subject_index="", para=Parameters()):
        TimeSeries.__init__(self, label="",ts=np.array([]),subject_index="", para=Parameters())
        self.label           = label
        self.subject_index   = subject_index
        self.timeseries      = ts
        self.shape           = ts.shape
        self.parameters      = deepcopy(para)
        self.BOLD            = TimeSeries()
        self.PARA            = np.array([])
        self.event_bold      = np.array([])
    
    # setters
    def set_para(self, PARA):
        self.PARA = PARA 

    def set_BOLD(self, BOLD):
        self.BOLD = BOLD 

    def set_event_bold(self, event_bold):
        self.event_bold = event_bold 
    
    # getters
    def get_event_bold(self):
        return self.event_bold

    def get_associated_BOLD(self):
        return self.BOLD 

    def get_HRF_para(self):
        return self.PARA 

    # misc.
    def compareTimeSeries(self, ts):
        """ Compares whether another HRF time-series is similar to it.
            Three checks are performed for this:
                1. Label
                2. rsHRF Parameters
                3. Preprocessed-BOLD which was used as input
            If all the three comparisions return true, then both the HRF
                time-series objects are identical 
        """
        if self.label == ts.get_label() and self.parameters.compareParameters(ts.get_parameters()) and self.BOLD.compareTimeSeries(ts.get_associated_BOLD()):
            return True 
        else:
            return False

    def save_info(self, name):
        """ Saves the information about the time-series in a .mat file """
        try:
            dic = {}
            dic["timeseries"] = self.timeseries
            dic["PARA"] = self.PARA 
            dic["eventBold"] = self.event_bold
            para = self.parameters.get_parameters()
            for each in para.keys():
                dic[each] = para[each] 
            savemat(name, dic)
            return True
        except:
            return False


import numpy as np

class Subject():
    """
    Stores the information corresponding to a particular subject.

    Attrbutes:
        1. index          : This is the index of the subject (as determined by BIDS convention)
        2. BOLD_raw       : An array which stores the corresponding Raw BOLD time-series for the subject
        3. BOLD_pre       : An array which stores the corresponding Preprocessed-BOLD time-series for the subject
        4. BOLD_deconv    : An array which stores the corresponding Deconvolved-BOLD time-series for the subject
        5. HRF            : An array which stores the corresponding Hemodynamic Response Function time-series for the subject

        -> All the attributes from 2-5, are arrays of TimeSeries objects
    """
    def __init__(self, index):
        self.index          = index
        self.BOLD_raw       = []
        self.BOLD_pre       = []
        self.BOLD_deconv    = []
        self.HRF            = []
    
    # getters
    def get_input_filename(self):
        return self.input_filename 
    
    def get_subject_index(self):
        return self.index
    
    def get_BOLD_raw(self):
        return tuple(self.BOLD_raw)
    
    def get_BOLD_pre(self):
        return tuple(self.BOLD_pre)
    
    def get_BOLD_deconv(self):
        return tuple(self.BOLD_deconv)
    
    def get_HRF(self):
        return tuple(self.HRF)

    # adding to time-series objects of the existing subject
    def add_BOLD_raw(self, ts):
        self.BOLD_raw.append(ts) 
        return len(self.BOLD_raw) - 1

    def add_BOLD_deconv(self, ts):
        self.BOLD_deconv.append(ts)
        return len(self.BOLD_raw) - 1

    def add_BOLD_pre(self, ts):
        self.BOLD_pre.append(ts)
        return len(self.BOLD_raw) - 1

    def add_HRF(self, ts):
        self.HRF.append(ts)
        return len(self.BOLD_raw) - 1

    # misc.
    def is_present(self, label, misc, getts=False):
        """ Checks whether a time-series is already present 
            Misc takes in all the relevant information which determines the uniqueness
                of a time-series
            If getts = True, the function returns the time-series object if it is present """
        if label == "BOLD":
            # Looking for Raw BOLD Data
            for each in self.BOLD_raw:
                # Determines whether the Raw BOLD data is already present
                # Checks for the input-file
                if misc == each.get_inputfile():
                    if getts :
                        return each
                    return True 
        elif label == "Preprocessed-BOLD":
            # Looking for Preprocessed BOLD Data
            para     = misc[0]
            mask     = misc[1]
            bold     = misc[2]
            for each in self.BOLD_pre:
                # Determines whether Preprocessed BOLD data is already present
                # Checks the parameters, mask-file and RAW Bold
                if para.compareParameters(each.get_parameters()) \
                    and each.get_maskfile() == misc[1] \
                    and bold.compareTimeSeries(each.get_BOLD_Raw()):
                    if getts:
                        return each
                    return True 
        elif label == "HRF":
            # Looking for HRF Data
            para     = misc[0]
            BOLD_pre = misc[1]
            for each in self.HRF:
                # Determines whether the HRF is already present
                # Checks the parameters and Preprocessed BOLD
                if para.compareParameters(each.get_parameters()) \
                    and BOLD_pre.compareTimeSeries(each.get_associated_BOLD()):
                    if getts:
                        return each
                    return True
        elif label == "Deconvolved-BOLD":
            # Looking for Deconvolved BOLD Data
            para = misc[0]
            HRF  = misc[1]
            for each in self.BOLD_deconv:
                # Determines whether the Deconvolved BOLD is already present
                # Checks the associated HRF
                if para.compareParameters(each.get_parameters()) \
                    and HRF.compareTimeSeries(each.get_associated_HRF()):
                    if getts :
                        return each
                    return True
        return False
    
    def get_time_series_pos(self, ts):
        """ 
        Takes the time-series as input and returns its position in the array
        """
        label = ts.get_label()
        if label    == "BOLD":
            arr     = self.BOLD_raw 
        elif label  == "Preprocessed-BOLD":
            arr     = self.BOLD_pre 
        elif label  == "Deconvolved-BOLD":
            arr     = self.BOLD_deconv
        elif label  == "HRF":
            arr     = self.HRF 
        else :
            arr     = []
        for i in range(len(arr)):
            if ts.compareTimeSeries(arr[i]):
                return str(i)
        return None

    def get_time_series_by_index(self, ts_type, index):
        """ Takes the index of a time-series and returns the time-series """
        if ts_type == "BOLD":
            arr = self.BOLD_raw 
        elif ts_type == "Preprocessed-BOLD":
            arr = self.BOLD_pre 
        elif ts_type == "Deconvolved-BOLD":
            arr = self.BOLD_deconv 
        elif ts_type == "HRF":
            arr = self.HRF 
        else:
            return
        return arr[index]

    def get_plotables(self):
        """ 
        Returns an array of all the time-series objects that can be plotted for the subject
        The array contains of tuples of the format : (time-series labels, time-series numpy arrays) 
        """
        out = []
        for i in range(len(self.BOLD_raw)):
            out.append((self.index+"_BOLD_"+str(i),self.BOLD_raw[i].get_ts()))
        for i in range(len(self.BOLD_pre)):
            out.append((self.index+"_Preprocessed-BOLD_"+str(i),self.BOLD_pre[i].get_ts()))
        for i in range(len(self.BOLD_deconv)):
            out.append((self.index+"_Deconvolved-BOLD_"+str(i),self.BOLD_deconv[i].get_ts()))
        for i in range(len(self.HRF)):
            out.append((self.index+"_HRF_"+str(i),self.HRF[i].get_ts()))
        return out

    def get_data_labels(self):
        """
        Returns an array with labels for all the time-series objects for the subject 
        """
        out = []
        for i in range(len(self.BOLD_raw)):
            out.append(self.index+"_BOLD_"+str(i))
        for i in range(len(self.BOLD_pre)):
            out.append(self.index+"_Preprocessed-BOLD_"+str(i))
        for i in range(len(self.BOLD_deconv)):
            out.append(self.index+"_Deconvolved-BOLD_"+str(i))
        for i in range(len(self.HRF)):
            out.append(self.index+"_HRF_"+str(i))
        return out

    
    



import os
import numpy    as np
import nibabel  as nib
import scipy.io as sio
from bids.grabbids  import BIDSLayout
from scipy          import stats, signal
from scipy.sparse   import lil_matrix
from ...            import spm_dep, processing, parameters, basis_functions, utils

from ..datatypes.timeseries.hrf               import HRF
from ..datatypes.timeseries.bold_raw          import BOLD_Raw
from ..datatypes.timeseries.bold_preprocessed import BOLD_Preprocessed
from ..datatypes.timeseries.bold_deconv       import Bold_Deconv
from ..datatypes.misc.parameters              import Parameters
from ..datatypes.misc.subject                 import Subject
from ..datatypes.misc.store                   import Store
from ..misc.status                            import Status

import warnings
warnings.filterwarnings("ignore")

class Core():
    """
    All of the processing occurs here
    """
    def __init__(self):
        # constructor
        self.parameters = Parameters() # rsHRF parameters
        self.dataStore  = Store()      # stores the information for the subjects

    # getters for data-store
    def updateParameters(self, dic={}):  
        # updates the rsHRF-parameters   
        return self.parameters.set_parameters(dic) 

    def get_parameters(self):
        # returns the rsHRF-parameters
        return self.parameters.get_parameters()    

    def get_time_series(self, curr):
        # gets the time-series
        return self.dataStore.get_time_series(curr)

    def get_plotables(self, curr):
        # gets all the time-series objects that can be plotted
        return self.dataStore.get_plotables(curr)

    def get_store_info(self, curr):
        # gets the information regarding the time-series objects
        return self.dataStore.get_info(curr)

    def get_subjects(self):
        # gets all the subjects present in the data-store
        return self.dataStore.get_subjects()
    
    def get_data_labels(self, curr):
        # gets all the labels present in the data-store
        return self.dataStore.get_data_labels(curr)

    # saving data
    def save_info(self, curr, out):
        return self.dataStore.save_info(curr, out)
    
    def makeInput(self, inp):
        """
        Obtains the input (file-paths and input-mode) from the main window
        and updates the subjects/time-series objects appropriately
        """
        # obtains the Raw BOLD data and preprocessed data, and updates the data-store
        input_file, mask_file, file_type, mode = inp # each input comprises of an input_file, mask_file and a file_type, and the mode 
        if mode == "file":                           # if the mode is file (that is input is a stand-alone NIfTI or GIfTI file)
            return self.makeInputFile(input_file, mask_file, file_type, 'file')
        # if the input is present in a directory organized in BIDS format
        elif "bids" in mode:
            # obtaining all the input files
            layout              = BIDSLayout(input_file)
            subjects_to_analyze = layout.get_subjects()
            # if no subjects were found
            if not subjects_to_analyze:
                return Status(False, error='Could not find participants. Please make sure the BIDS data '
                         'structure is present and correct. Datasets can be validated online '
                         'using the BIDS Validator (http://incf.github.io/bids-validator/).')
            all_inputs = layout.get(modality='func', subject=subjects_to_analyze, task='rest', type='preproc', extensions=['nii', 'nii.gz'])
            # if no *preproc.nii or *preproc.nii.gz files were found
            if not all_inputs != []:
                return Status(False, error = 'There are no files of type *preproc.nii / *preproc.nii.gz '
                         'Please make sure to have at least one file of the above type '
                         'in the BIDS specification')
            # if the atlas file is provided
            if mode == "bids w/ atlas":
                for file_count in range(len(all_inputs)):
                    # making input for every combination of input-file and atlas
                    temp = (self.makeInputFile(all_inputs[file_count], mask_file, ".nii", 'bids w/ atlas'))
                    if not temp.get_state():
                        return temp 
                # returns the status
                return Status(True, info="Preprocessed all input", dic={"Number of Input Subjects: ":len(all_inputs)})
            # if the atlas file is not provided
            elif mode == "bids":
                # obtaining the masks
                all_masks = layout.get(modality='func', subject=subjects_to_analyze, task='rest', type='brainmask', extensions=['nii', 'nii.gz'])
                # if no mask files were found
                if not all_masks != []:
                    return Status(False, error='There are no files of type *brainmask.nii / *brainmask.nii.gz '
                             'Please make sure to have at least one file of the above type '
                             'in the BIDS specification')
                # if the number of mask files != number of input files
                if len(all_inputs) != len(all_masks):
                    return Status(False, error='The number of *preproc.nii / .nii.gz and the number of '
                                 '*brainmask.nii / .nii.gz are different. Please make sure that '
                                 'there is one mask for each input_file present')
                all_inputs.sort()
                all_masks.sort()
                # matching the prefix to align the inputs and the masks together
                all_prefix_match   = False
                prefix_match_count = 0
                for i in range(len(all_inputs)):
                    input_prefix    = all_inputs[i].filename.split('/')[-1].split('_preproc')[0]
                    mask_prefix     = all_masks[i].filename.split('/')[-1].split('_brainmask')[0]
                    if input_prefix == mask_prefix:
                        prefix_match_count += 1
                    else:
                        all_prefix_match = False
                        break
                if prefix_match_count == len(all_inputs):
                    all_prefix_match  = True
                # if the mask and input files have different prefix
                if not all_prefix_match:
                    return Status(False, error='The mask and input files should have the same prefix for correspondence. '
                                 'Please consider renaming your files')
                else:
                    # making input for every combination of input-file and mask-file
                    for file_count in range(len(all_inputs)):
                        temp = (self.makeInputFile(all_inputs[file_count], all_masks[file_count], ".nii", 'bids'))
                        if not temp.get_state():
                            return temp 
                    # returns the status
                    return Status(True, info="Preprocessed all input", dic={"Number of Input Subjects: ":len(all_inputs)})

    def makeInputFile(self, input_file, mask_file, file_type, mode):
        """
        Obtains the Raw BOLD Time-series and Preprocessed BOLD Time-series
        when the input is a stand-alone NIfTI or GIfTI file
        """
        # if the input is in form of a File object
        if 'bids' in mode:
            input_file = input_file.filename
        # if the mask is in form of a File object
        if mode == 'bids':
            mask_file = mask_file.filename
        # getting the subject index
        try:
            subject_index =  input_file.split("/")[-1].split("_")[0][4:]
        except:
            return Status(False, error="Input file should begin with 'sub-'")
        # obtaining the header for the mask
        try:  
            v    = spm_dep.spm.spm_vol(mask_file)
        except :
            return Status(False, error="Invalid Mask File!")
        # obtaining the header for the input
        try:   
            v1   = spm_dep.spm.spm_vol(input_file)
        except: 
            return Status(False, error="Invalid Input File!")
        # obtaining the mask data
        if file_type   == ".nii" or file_type == ".nii.gz":
            brain       = spm_dep.spm.spm_read_vols(v)
        elif file_type == ".gii" or file_type == ".gii.gz":
            brain       = v.agg_data().flatten(order='F')
        else:
            return Status(False, error="Invalid Input File Type!")
        # brain voxel indices with as obtained by mask-data
        voxel_ind = np.where(brain > 0)[0]
        # checking the dimensions of the mask-file and input-file
        if ((file_type == ".nii" or file_type == ".nii.gz") and v1.header.get_data_shape()[:-1] != v.header.get_data_shape()) or ((file_type == ".gii" or file_type == ".gii.gz") and v.agg_data().shape[0]!= v.agg_data().shape[0]):
            return Status(False, error='The dimension of your mask is different than the one of your fMRI data!')
        # checking whether the subject is already present in the datastore
        if not self.dataStore.is_present(subject_index):
            # instantiating a new object if it is not present
            subject = Subject(index=subject_index)
            self.dataStore.add_subject(subject)
        else:
            # obtaining the object if it is present
            subject = self.dataStore.get_subject_by_index(subject_index)
        # checks if the BOLD time-series is already present
        if not subject.is_present("BOLD", input_file):   # if the time-series is not present
            if file_type   == ".nii" or file_type == ".nii.gz" :
                data            = v1.get_data()                 
                nobs            = data.shape[3]
                data1           = np.reshape(data, (-1, nobs), order='F').T
            elif file_type == ".gii" or file_type == ".gii.gz":
                data            = v1.agg_data()
                _, nobs         = data.shape
                data1           = np.reshape(data, (-1, nobs), order='F').T    
            bold_ts             = BOLD_Raw(label="BOLD", ts=np.array(data1), para=self.parameters, subject_index=subject.get_subject_index())   
            bold_ts.set_inputfile(input_file)   # the Raw BOLD Time-series is associated to a particular input-file 
            subject.add_BOLD_raw(bold_ts)       # adding the time-series object
        else:
            bold_ts             = subject.is_present("BOLD", input_file, getts=True)
        data1                   = bold_ts.get_ts()
        # checks if the Preprocessed-BOLD time-series is already present
        if not subject.is_present("Preprocessed-BOLD", (self.parameters, mask_file, bold_ts)):
            # pre-processing the time-series
            bold_sig = stats.zscore(data1[:, voxel_ind], ddof=1)
            bold_sig = np.nan_to_num(bold_sig)
            # instantiating the time-series objects
            bold_pre_ts = BOLD_Preprocessed(label="Preprocessed-BOLD", ts=np.array(bold_sig), para=self.parameters, subject_index=subject.get_subject_index())
            bold_pre_ts.set_BOLD_Raw(bold_ts)   # the Preprocessed BOLD Time-series is associated to a particular Raw BOLD Time-Series
            bold_pre_ts.set_maskfile(mask_file) # and a particular mask_file
            subject.add_BOLD_pre(bold_pre_ts)   # adding the time-series objects
            # returning the status
            return Status(True, info="Preprocessed Input!", dic={"Number of Time-Slices: ": bold_ts.get_shape()[0], "Number of Brain-Voxels: ":bold_ts.get_shape()[1]})
        return Status(False, error="Time series already exists. No new additions were made.")

    def retrieveHRF(self, bold_pre_ts, get_pos=False):
        """
        Retrieves the resting-state Hemodynamic Response Function
        with Preprocessed BOLD Time-series as input, and self.parameters
        as the parameters
        """
        subject_index = bold_pre_ts.get_subject_index()                    # gets the subject-index associated to the preprocessed BOLD Time-series
        subject       = self.dataStore.get_subject_by_index(subject_index) # gets the subject from the index
        # if the HRF has already been retrieved for this particular set of inputs
        if not subject.is_present("HRF", (self.parameters, bold_pre_ts)): 
            # inputs for retrieving the HRF
            bold_sig  = bold_pre_ts.get_ts()
            bold_sig  = processing. \
                        rest_filter. \
                        rest_IdealFilter(bold_sig, self.parameters.get_TR(), self.parameters.get_passband_deconvolve())
            para      = self.parameters.get_parameters()
            if not (para['estimation'] == 'sFIR' or para['estimation'] == 'FIR'):
                # estimate HRF for the fourier / hanning / gamma / cannon basis functions
                bf = basis_functions.basis_functions.get_basis_function(bold_sig.shape, para)       # obtaining the basis set
                beta_hrf, event_bold = utils.hrf_estimation.compute_hrf(bold_sig, para, [], -1, bf)
                hrfa = np.dot(bf, beta_hrf[np.arange(0, bf.shape[1]), :])
            else:
                # estimate HRF for FIR and sFIR
                beta_hrf, event_bold = utils.hrf_estimation.compute_hrf(bold_sig, para, [], -1)
                hrfa = beta_hrf
            # instantiating the time-series objects
            hrf = HRF(label="HRF", ts=hrfa, subject_index=subject_index, para=self.parameters)
            hrf.set_BOLD(bold_pre_ts)      # the HRF is associated to a particular Preprocessed BOLD Time-series
            hrf.set_event_bold(event_bold) # setting the bold-events
            pos = subject.add_HRF(hrf)     # adding the HRF to the subject
            # returning the status
            if get_pos:
                return Status(True, info="Retrieved HRF!", dic={"Shape of HRF: ": hrf.get_shape()}), pos
            else:
                return  Status(True, info="Retrieved HRF!", dic={"Shape of HRF: ": hrf.get_shape()})
        if get_pos:
            return Status(False, error="Time series already exists. No new additions were made."), None
        else:
            return Status(False, error="Time series already exists. No new additions were made.")

    def getHRFParameters(self, hrf):
        """
        Retrieves the resting-state Hemodynamic Response Function Parameters
        with HRF as input, and self.parameters as the parameters
        """
        # checking if the parameters have already been computed for HRF
        if hrf.get_HRF_para().size != 0:
            return Status(False,error="Parameters have already been computed for this HRF")
        # inputs for obtaining the parameters
        hrfa    = hrf.get_ts()
        para    = hrf.get_parameters().get_parameters()
        nvar    = hrfa.shape[1]
        PARA    = np.zeros((3, nvar))
        average = [0 for i in range(3)]
        # obtaining the parameters for each voxel
        for voxel_id in range(nvar):
            hrf1              = hrfa[:, voxel_id]
            PARA[:, voxel_id] = parameters.wgr_get_parameters(hrf1, para['TR'] / para['T'])
            for i in range(3):
                average[i] += PARA[i, voxel_id] 
        for i in range(3):
            average[i] /= nvar
        # setting the parameters
        hrf.set_para(PARA)
        # returning the status
        return Status(True, info="Retrieved HRF Parameters", dic={"Average Height: ":str(average[0])[:3], \
         "Avearage Time-To-Peak: ": str(average[1])[:3] + " seconds", "Avearge Full-Width at Half-Max: ": str(average[2])[:3]})

    def deconvolveHRF(self, hrf):
        """
        Retrieves the Deconvolved BOLD Time-series
        with HRF as input, and self.parameters as the parameters
        """
        subject_index = hrf.get_subject_index()                                     # gets the subject-index associated to the preprocessed BOLD Time-series
        subject       = self.dataStore.get_subject_by_index(subject_index)          # gets the subject from the index
        para          = self.parameters.get_parameters()    
        # if the HRF has already been retrieved for this particular set of inputs
        if not subject.is_present("Deconvolved-BOLD", (self.parameters, hrf)): 
            # inputs for retrieving the deconvolved BOLD
            hrfa              = hrf.get_ts()
            bold_sig          = hrf.get_associated_BOLD().get_ts()
            bold_sig          = processing. \
                                rest_filter. \
                                rest_IdealFilter(bold_sig, self.parameters.get_TR(), self.parameters.get_passband_deconvolve())
            event_bold        = hrf.get_event_bold()
            nvar              = hrfa.shape[1]
            nobs              = bold_sig.shape[0]
            data_deconv       = np.zeros(bold_sig.shape)
            event_number      = np.zeros((1, bold_sig.shape[1]))
            if para['T'] > 1:
                hrfa_TR = signal.resample_poly(hrfa, 1, para['T'])
            else:
                hrfa_TR = hrfa
            # obtaining the deconvolved BOLD for each voxel
            for voxel_id in range(nvar):
                hrf_                       = hrfa_TR[:, voxel_id]
                H                          = np.fft.fft(np.append(hrf_, np.zeros((nobs - max(hrf_.shape), 1))), axis=0)
                M                          = np.fft.fft(bold_sig[:, voxel_id])
                data_deconv[:, voxel_id]   = np.fft.ifft(H.conj() * M / (H * H.conj() + .1*np.mean((H * H.conj()))))
                event_number[:, voxel_id]  = np.amax(event_bold[voxel_id].shape)
            # instantiating the time-series object
            dd = Bold_Deconv(label="Deconvolved-BOLD", ts=data_deconv,subject_index=subject_index, para=hrf.get_parameters())
            dd.set_HRF(hrf) # the deconvoled BOLD Time-series is associated to a particular HRF
            dd.set_event_num(event_number) # setting the BOLD events
            pos = subject.add_BOLD_deconv(dd) # adding the time-series to the subject
            # returning the status
            return Status(True, info="Deconvolved BOLD Signal")
        return Status(False, error="Time series already exists. No new additions were made.")


import numpy as np
from scipy.io   import savemat
from copy       import deepcopy

from ...datatypes.misc.parameters import Parameters 
from .timeseries                  import TimeSeries

class BOLD_Raw(TimeSeries):
    """
    This stores the Raw BOLD Time-series
    
    Attributes:
        1. input_file : the input-file path of the Raw BOLD Time-series
    """
    def __init__(self, label="",ts=np.array([]),subject_index="", para=Parameters()):
        TimeSeries.__init__(self, label="",ts=np.array([]),subject_index="", para=Parameters())
        self.label          = label 
        self.subject_index  = subject_index 
        self.timeseries     = ts 
        self.shape          = ts.shape
        self.parameters     = deepcopy(para) 
        self.input_file     = ""

    # setters
    def set_inputfile(self, input_file):
        self.input_file = input_file

    # getters
    def get_inputfile(self):
        return self.input_file

    # misc.
    def compareTimeSeries(self, ts):
        """ Compares another time-series with itself to determine if both are identical 
            Two checks are performed:
                1. Label
                2. Input-file name
            If all the three comparisions return true, then both the HRF
                time-series objects are identical
        """
        if self.label == ts.get_label() and self.input_file == ts.get_inputfile():
            return True 
        else:
            return False 

    def save_info(self, name):
        """ Saves the information about the time-series in a .mat file """
        try:
            dic = {}
            dic["timeseries"] = self.timeseries
            dic["input_file"] = self.input_file 
            savemat(name, dic)
            return True
        except:
            return False
